#include "framecpp/config.h"

extern "C" {

#include <strings.h>

} // extern "C"

#include <algorithm>
#include <iterator>
#include <memory>
#include <sstream>
#include <stdexcept>
#include <string>

#include "general/util.hh"
#include "general/autoarray.hh"
#include "general/types.hh"

#include "framecpp/Common/Compression.hh"
#include "framecpp/Common/CompressionDifferential.hh"
#include "framecpp/Common/CompressionGZip.hh"
#include "framecpp/Common/CompressionZeroSuppress.hh"

using namespace FrameCPP;
using General::AutoArray;

#define LM_INFO 0

#if LM_DEBUG
#define AT( ) std::cerr << "INFO: " << __FILE__ << " " << __LINE__ << std::endl;
#else
#define AT( )
#endif

namespace
{
  typedef General::hash_map< INT_4U, INT_4U > data_type_size_mapping;

  using namespace FrameCPP::Compression;

  data_type_size_mapping& size_map_init( )
  {
    static data_type_size_mapping m;

#define MAP( A, B) \
    m[ A ] = sizeof( B )

    MAP( FR_VECT_C, CHAR );
    MAP( FR_VECT_2S, INT_2S );
    MAP( FR_VECT_8R, REAL_8 );
    MAP( FR_VECT_4R, REAL_4 );
    MAP( FR_VECT_4S, INT_4S );
    MAP( FR_VECT_8S, INT_8S );
    MAP( FR_VECT_8C, COMPLEX_8 );
    MAP( FR_VECT_16C, COMPLEX_16 );
    MAP( FR_VECT_2U, INT_2U );
    MAP( FR_VECT_4U, INT_4U );
    MAP( FR_VECT_8U, INT_8U );
    MAP( FR_VECT_1U, CHAR_U );

#undef MAP
  
    return m;
  }

  static const data_type_size_mapping&	size_map = size_map_init( );

  template < typename _Map >
  inline typename _Map::key_type
  reverse_lookup( typename _Map::mapped_type Key,
		  const _Map& Map )
  {
    for ( typename _Map::const_iterator
	    cur = Map.begin( ),
	    last = Map.end( );
	  cur != last;
	  ++cur )
    {
      if ( cur->second == Key )
      {
	return cur->first;
      }
    }
    
    std::ostringstream	msg;

    msg << "Cannot reverse map: " << Key;
    throw std::range_error( msg.str( ) );
  }

  INT_2U
  data_size( INT_4U Type )
  {
    data_type_size_mapping::const_iterator cur = size_map.find( Type );

    if ( cur == size_map.end( ) )
    {
      std::ostringstream	msg;

      msg << "unknown size for data type: " << Type;

      throw std::range_error( msg.str( ) );
    }
    return cur->second;
  }

  template < class T >
  inline std::string dump( const void* Input, INT_8U NData )
  {
    const T* input( reinterpret_cast< const T* >( Input ) );
    std::ostringstream	oss;

    std::copy( input, &( input[ NData ] ),
	       std::ostream_iterator< const T >( oss, " ") );
	  
    return oss.str( );
  }

  template < >
  inline std::string dump< CHAR_U >( const void* Input, INT_8U NData )
  {
    const CHAR_U* input( reinterpret_cast< const CHAR_U* >( Input ) );
    std::ostringstream	oss;

    std::copy( input, &( input[ NData ] ),
	       std::ostream_iterator< const int >( oss, " ") );
	  
    return oss.str( );
  }

  void swap_data( INT_8U NData,
		  FrameCPP::Compression::data_types_type Type,
		  CHAR_U* Buffer );

  INT_4U
  DiffGZip( const CHAR_U* Source,
	    const INT_8U SourceLength,
	    const INT_8U SourceByteLength,
	    INT_4U DataType,
	    INT_4U Level,
	    const compress_type_mapping& CompressionMapping,
	    AutoArray< CHAR_U>& Dest,
	    INT_8U& DestByteLength )
  {
    register compress_type t( MODE_RAW );

    AutoArray< CHAR_U >	diff_buffer;

    switch( DataType )
    {
    case FR_VECT_C:
    case FR_VECT_1U:
    case FR_VECT_2S:
    case FR_VECT_2U:
    case FR_VECT_4S:
    case FR_VECT_4U:
      Differential::Encode( Source, SourceLength, DataType, diff_buffer );
      GZip::Compress( diff_buffer.get( ), SourceByteLength, Level,
		      Dest, DestByteLength );
      // reset_data_pointer( Data, DataByteLength );
      t = MODE_DIFF_GZIP;
      break;
    default:
      GZip::Compress( Source, SourceByteLength, Level,
		      Dest, DestByteLength );
      t = MODE_GZIP;
      break;
    }
    return (INT_4U)reverse_lookup( t, CompressionMapping );
  }

} // namespace anonymous

namespace FrameCPP
{
  namespace Compression
  {
    //-------------------------------------------------------------------
    /// This will take on input buffer and generate a compressed version
    /// of the input buffer.
    //-------------------------------------------------------------------
    void
    Compress( INT_4U& CompressionMode, INT_4U Level,
	      const compress_type_mapping& CompressionMapping,
	      INT_4U DataType,
	      const data_type_mapping& DataTypeMapping,
	      const CHAR_U* Source,
	      const INT_8U SourceLength,
	      const INT_8U SourceByteLength,
	      CHAR_U*& Dest,
	      INT_8U& DestByteLength )
    {
      //-----------------------------------------------------------------
      // Check for the very basic case of no compression
      //-----------------------------------------------------------------
      if ( CompressionMode == MODE_RAW )
      {
	return;
      }
      //-----------------------------------------------------------------
      // Setup for failure.
      //-----------------------------------------------------------------
      Dest = (CHAR_U*)NULL;
      DestByteLength = 0;
      General::AutoArray< CHAR_U >	tmp;
      General::AutoArray< CHAR_U >	data_out;
      INT_8U				data_out_len = 0;	// bytes

      //-----------------------------------------------------------------
      //-----------------------------------------------------------------
      INT_4U compression_mode_return = CompressionMode;
      {
	//---------------------------------------------------------------
	// Translate the compression mode
	//---------------------------------------------------------------
	compress_type_mapping::const_iterator c
	  = CompressionMapping.find( CompressionMode );

	if ( c == CompressionMapping.end( ) )
	{
	  std::ostringstream	msg;

	  msg << "compression: unknown compression mode: " << CompressionMode;
	  throw std::range_error( msg.str( ) );
	}
	CompressionMode = c->second;
      }
      {
	//---------------------------------------------------------------
	// Translate the data type
	//---------------------------------------------------------------
	data_type_mapping::const_iterator d
	  = DataTypeMapping.find( DataType );

	if ( d == DataTypeMapping.end( ) )
	{
	  std::ostringstream	msg;

	  msg << "compression: unknown data type: " << DataType;
	  throw std::range_error( msg.str( ) );
	}
	DataType = d->second;
      }

      //-----------------------------------------------------------------
      // Do the compression
      //-----------------------------------------------------------------
      AT( );
      bool gzip_others( false );

      switch( CompressionMode )
      {
      case MODE_GZIP:
	//---------------------------------------------------------------
	// Compress using GZip algorithm
	//---------------------------------------------------------------
	AT( );
	GZip::Compress( Source, SourceByteLength, Level,
			data_out, data_out_len );
	break;
      case MODE_DIFF:
	Differential::Encode( Source, SourceLength, DataType, data_out );
	data_out_len = SourceByteLength;
	break;
      case MODE_DIFF_GZIP:
	AT( );
	compression_mode_return
	  = DiffGZip( Source, SourceLength, SourceByteLength, DataType,
		      Level, CompressionMapping,
		      data_out, data_out_len );
	break;
      case MODE_ZERO_SUPPRESS_OTHERWISE_GZIP:
      case MODE_ZERO_SUPPRESS_SHORT_GZIP_OTHER:
	gzip_others = true;
	// :TRICKY: Fall through
      case MODE_ZERO_SUPPRESS_INT_FLOAT:
      case MODE_ZERO_SUPPRESS_SHORT:
	{
	  AutoArray< CHAR_U >	diff_buffer;
	  //-------------------------------------------------------------
	  /// \todo See about doing gzip compression when an exception
	  ///       is thrown for zero suppress compressions.
	  //-------------------------------------------------------------
	  switch( DataType )
	  {
	  case FR_VECT_2S:
	  case FR_VECT_2U:
	    //-------------------------------------------------------------
	    // All modes support at least this
	    //-------------------------------------------------------------
	    ZeroSuppress::DiffCompress< 2 >( Source, SourceLength, DataType,
					     data_out, data_out_len );
	    if ( data_out_len )
	    {
	      compression_mode_return 
		= INT_4U( reverse_lookup( MODE_ZERO_SUPPRESS_SHORT,
					  CompressionMapping ) );
	      gzip_others = false;
	    }
	    break; // FR_VECT_2S; FR_VECT_2U

	  case FR_VECT_4R:
	  case FR_VECT_4S:
	  case FR_VECT_4U:
	    if ( ( MODE_ZERO_SUPPRESS_OTHERWISE_GZIP == CompressionMode )
		 || ( MODE_ZERO_SUPPRESS_INT_FLOAT == CompressionMode ) )
	    {
	      //-----------------------------------------------------------
	      // Compress the buffer
	      //-----------------------------------------------------------
	      ZeroSuppress::DiffCompress< 4 >( Source, SourceLength, DataType,
					       data_out, data_out_len );
	      if ( data_out_len )
	      {
		compression_mode_return
		  = INT_4U( reverse_lookup( MODE_ZERO_SUPPRESS_INT_FLOAT,
					    CompressionMapping ) );
		//---------------------------------------------------------
		// Flag buffer as compressed.
		//---------------------------------------------------------
		gzip_others = false;
	      }
	    }
	    break; // FR_VECT_4R; FR_VECT_4S; FR_VECT_4U
	  }
	  if ( gzip_others )
	  {
	    GZip::Compress( Source, SourceByteLength, Level,
			    data_out, data_out_len );
	    if ( data_out_len )
	    {
	      compression_mode_return
		= (INT_4U)reverse_lookup( MODE_GZIP,
					  CompressionMapping );
	    }
	  }
	}
	break;
      default:
	//---------------------------------------------------------------
	// Have not been able to do compression
	//---------------------------------------------------------------
	{
	  std::ostringstream	msg;

	  msg << "unsupported compression mode: " << CompressionMode;
	  throw std::range_error( msg.str( ) );
	}
	break;
      }

      //-----------------------------------------------------------------
      // prepare for return
      //-----------------------------------------------------------------
      if ( data_out_len == 0 )
      {
	data_out.reset( (CHAR_U*)NULL );
	compression_mode_return = MODE_RAW;
      }
      CompressionMode = compression_mode_return;
      Dest = data_out.release( );
      DestByteLength = data_out_len;
    }

    void
    Expand( const bool NativeOrder,
	    const INT_4U CompressionMode,
	    const compress_type_mapping& CompressionMapping,
	    const INT_4U DataType,
	    const data_type_mapping& DataTypeMapping,
	    const CHAR_U* Source,
	    const INT_8U SourceLength,
	    const INT_8U SourceByteLength,
	    CHAR_U*& Dest,
	    INT_8U& DataByteLength )
    {
      INT_4U 		cm( 0 );
      data_types_type	dt( data_types_type( 0 ) );

      {
	//---------------------------------------------------------------
	// Translate the compression mode
	//---------------------------------------------------------------
	compress_type_mapping::const_iterator c
	  = CompressionMapping.find( CompressionMode );

	if ( c == CompressionMapping.end( ) )
	{
	  std::ostringstream	msg;

	  msg << "unknown compression mode: " << CompressionMode;
	  throw std::range_error( msg.str( ) );
	}
	cm = c->second;
      }
      AT( );
      //----------------------------------------------------------------
      // Check if any work needs to happen
      //----------------------------------------------------------------
      if( ( CompressionMode == MODE_RAW ) && NativeOrder )
      {
	//---------------------------------------------------------------
	// Nothing to do since the source buffer is already uncompressed.
	//---------------------------------------------------------------
	return;
      }
      //-----------------------------------------------------------------
      // Prepare the output buffer since some translation needs to happen
      // to expand the source buffer.
      //-----------------------------------------------------------------
      {
	//---------------------------------------------------------------
	// Translate the data type
	//---------------------------------------------------------------
	data_type_mapping::const_iterator d
	  = DataTypeMapping.find( DataType );

	if ( d == DataTypeMapping.end( ) )
	{
	  std::ostringstream	msg;

	  msg << "unknown compression mode: " << DataType;
	  throw std::range_error( msg.str( ) );
	}
	dt = d->second;
      }

      //-----------------------------------------------------------------
      // Expand data
      //-----------------------------------------------------------------
      AutoArray< CHAR_U >	data_out;
      INT_8U			data_out_len = 0;

      switch( cm )
      {
      case MODE_RAW:
	if ( ! NativeOrder )
	{
	  data_out_len = SourceByteLength;
	  data_out.reset( new CHAR_U[ SourceByteLength ] );
	  std::copy( Source, &(Source[ SourceByteLength ]), data_out.get( ) );
	  swap_data( SourceLength, dt, data_out.get( ) );
	}
	break;
      case MODE_GZIP:
	//---------------------------------------------------------------
	// Compress using GZip algorithm
	//---------------------------------------------------------------
	AT( );
	data_out_len = data_size( dt ) * SourceLength;
	GZip::Expand( Source, SourceByteLength, data_out, data_out_len );
	if ( ! NativeOrder )
	{
	  swap_data( SourceLength, dt, data_out.get( ) );
	}
	break;
      case MODE_DIFF:
	{
	  INT_4U ndata( SourceLength );

	  if ( NativeOrder )
	  {
	    Differential::Decode( Source, ndata, dt, data_out );
	  }
	  else
	  {
	    data_out_len = SourceByteLength;
	    data_out.reset( new CHAR_U[ SourceByteLength ] );
	    std::copy( Source, &(Source[ SourceByteLength ]),
		       data_out.get( ) );
	    swap_data( SourceLength, dt, data_out.get( ) );
	    
	    Differential::Decode( data_out.get( ), ndata, dt );
	  }
	  data_out_len = SourceByteLength;
	}
	break;
      case MODE_DIFF_GZIP:
	{
	  AT( );
	  INT_4U ndata( SourceLength );

	  //-------------------------------------------------------------
	  // Do the gzip decompression
	  //-------------------------------------------------------------
	  data_out_len = data_size( dt ) * SourceLength;
	  GZip::Expand( Source, SourceByteLength, data_out, data_out_len );
	  //-------------------------------------------------------------
	  // Put the bytes into host order
	  //-------------------------------------------------------------
	  if ( ! NativeOrder )
	  {
	    swap_data( SourceLength, dt, data_out.get( ) );
	  }
	  //-------------------------------------------------------------
	  // Undo the differentiation in place.
	  //-------------------------------------------------------------
	  Differential::Decode( data_out.get( ), ndata, dt );
	}
	break;
      case MODE_ZERO_SUPPRESS_SHORT:
	{
	  AT( );
	  ZeroSuppress::DiffExpand< 2 >( Source, SourceByteLength,
					 !NativeOrder, dt,
					 data_out, SourceLength,
					 data_out_len );
	}
	break;
      case MODE_ZERO_SUPPRESS_INT_FLOAT:
	{
	  AT( );
	  ZeroSuppress::DiffExpand< 4 >( Source, SourceByteLength,
					 !NativeOrder, dt,
					 data_out, SourceLength,
					 data_out_len );
	}
	break;
      default:
	throw std::runtime_error( "FrameCPP::Compression::Expand is not implemented" );
	break;
      }
      //-----------------------------------------------------------------
      // Return the result to the caller
      //-----------------------------------------------------------------
      Dest = data_out.release( );
      DataByteLength = data_out_len;
    }
#if 0
    void
    Expand( bool NativeOrder,
	    INT_4U CompressionMode,
	    const compress_type_mapping& CompressionMapping,
	    INT_4U DataType,
	    const data_type_mapping& DataTypeMapping,
	    CHAR_U*& Data,
	    INT_8U& DataLength,
	    INT_8U& DataByteLength
	    )
    {
      AT( );
      // If it's not compressed, then it might need to be byte swapped   
      if( CompressionMode == MODE_RAW )
      {
	AT( );
	swap_data( DataLength, (data_types_type)DataType, Data );
	AT( );
	return;
      }

      AT( );
      // See if the data should be z-uncompressed (zero suppress method)
      // for short data only
      //
      if( ( CompressionMode == MODE_ZERO_SUPPRESS_SHORT_GZIP_OTHER ||
	    CompressionMode == MODE_ZERO_SUPPRESS_SHORT ) &&
	  ( DataType == FR_VECT_2S || DataType == FR_VECT_2U ) )
      {
	AT( );
	if ( ! NativeOrder )
	{
	  AT( );
	  swap_data( DataByteLength/sizeof( INT_2S ),
		     (data_types_type)DataType, Data );
	  NativeOrder = true;
	}

	ZeroSuppressUncompressShort( Data, DataByteLength, DataLength );
      }
      else if ( ( CompressionMode == MODE_ZERO_SUPPRESS_INT_FLOAT ) &&
		( DataType == FR_VECT_4S
		  || DataType == FR_VECT_4U
		  || DataType == FR_VECT_4R ) )
      {
	AT( );
	if ( ! NativeOrder )
	{
	  AT( );
	  swap_data( DataByteLength/sizeof( INT_4S ),
		     (data_types_type)DataType, Data );
	  NativeOrder = true;
	}
	ZeroSuppressUncompressInt( Data, DataByteLength, DataLength );
      }
      // Determine if we need to gunzip the data.
      if ( ( CompressionMode == MODE_GZIP ) ||
	   ( CompressionMode == MODE_DIFF_GZIP ) || 
	   ( CompressionMode == MODE_ZERO_SUPPRESS_SHORT_GZIP_OTHER &&
	     ( DataType != FR_VECT_2S
	       || DataType != FR_VECT_2U ) ) )
      {
	//---------------------------------------------------------------
	// Gunzip the buffer
	//---------------------------------------------------------------
	AT( );
	const INT_8U	length = DataLength * data_size( DataType );
	GZip::Expand( Data,		/* compressed data		*/
		      DataByteLength,	/* size of compressed data	*/
		      length		/* expected data length		*/
		      );
      }
    
      if ( ! NativeOrder )
      {
	AT( );
	swap_data( DataLength, (data_types_type)DataType, Data );
	NativeOrder = true;
      }

      // Integrate the data if necessary.
      if ( ( CompressionMode == MODE_DIFF ) ||
	   ( CompressionMode == MODE_DIFF_GZIP ) ||
	   ( CompressionMode == MODE_ZERO_SUPPRESS_SHORT_GZIP_OTHER ) ||
	   ( CompressionMode == MODE_ZERO_SUPPRESS_SHORT &&
	     ( DataType == FR_VECT_2S || DataType == FR_VECT_2U ) ) ||
	   ( CompressionMode == MODE_ZERO_SUPPRESS_INT_FLOAT &&
	     ( DataType == FR_VECT_4S ||
	       DataType == FR_VECT_4U ||
	       DataType == FR_VECT_4R ) ) )
      {
	Differential::Decode( Data, DataLength, (data_types_type)DataType );
      }
    } // func - Expand
#endif /* 0 */

  } // namespace - Compression
} // namespace - FrameCPP 

namespace
{
  void
  swap_data( INT_8U NData,
	     FrameCPP::Compression::data_types_type Type,
	     CHAR_U* Buffer )
  {
    using namespace FrameCPP::Compression;

    switch( Type )
    {
    case FR_VECT_2S:
    case FR_VECT_2U:
      reverse<2>( Buffer, NData );
      break;
    case FR_VECT_4R:
    case FR_VECT_4S:
    case FR_VECT_4U:
      reverse<4>( Buffer, NData );
      break;
    case FR_VECT_8R:
    case FR_VECT_8S:
    case FR_VECT_8U:
      reverse<8>( Buffer, NData );
      break;
    case FR_VECT_8C:
      reverse<4>( Buffer, NData * 2);
      break;
    case FR_VECT_16C:
      reverse<8>( Buffer, NData * 2);
      break;
    case FR_VECT_STRING:
      {
	CHAR_U*	pos( Buffer );
	INT_2U*	len( (INT_2U*)(Buffer ) );
	for ( INT_4U x = 0; x < NData; x++ )
	{
	  reverse< 2 >( pos, 1 );
	  pos += *len + 2;
	  len = (INT_2U*)pos;
	}
      }
      break;
    default:
      break;
    }
  }
} // namespace anonymous
