#ifndef FRAME_CPP__Common__Compression_hh
#define	FRAME_CPP__Common__Compression_hh

#include <limits.h>

#include <iostream>
#include <sstream>
#include <stdexcept>

#include "general/types.hh"
#include "general/autoarray.hh"
#include "general/hash_map.hh"

#include "framecpp/Common/CompressionException.hh"

namespace FrameCPP
{
  //---------------------------------------------------------------------
  /// \brief Routines related to compression algoriths
  //---------------------------------------------------------------------
  namespace Compression
  {
    //-------------------------------------------------------------------
    /// \brief Enumerated type for supported vector data types.
    //-------------------------------------------------------------------
    enum data_types_type {
      FR_VECT_C = 0,
      FR_VECT_2S = 1,
      FR_VECT_8R = 2,
      FR_VECT_4R = 3,
      FR_VECT_4S = 4,
      FR_VECT_8S = 5,
      FR_VECT_8C = 6,
      FR_VECT_16C = 7,
      FR_VECT_STRING = 8,
      FR_VECT_2U = 9,
      FR_VECT_4U = 10,
      FR_VECT_8U = 11,
      FR_VECT_1U = 12
    };

    //-------------------------------------------------------------------
    /// \brief Enumerated type for general compression modes
    //-------------------------------------------------------------------
    enum compress_type {
      MODE_RAW = 0,
      MODE_GZIP = 1,
      MODE_DIFF = 2,
      MODE_DIFF_GZIP = 3,
      MODE_ZERO_SUPPRESS_SHORT = 5,
      MODE_ZERO_SUPPRESS_SHORT_GZIP_OTHER = 6,
      MODE_ZERO_SUPPRESS_INT_FLOAT = 8,
      MODE_ZERO_SUPPRESS_OTHERWISE_GZIP = 10
    };

    //===================================================================
    // types
    //===================================================================
    typedef INT_2U compression_base_type;

    typedef General::hash_map< INT_4U, data_types_type > data_type_mapping;
    typedef General::hash_map< INT_4U, compress_type > compress_type_mapping;

    //-------------------------------------------------------------------
    /// \brief compress data without altering the source
    ///
    /// \param[in,out] CompressionMode
    ///     On input, this variable represents the desired compression
    ///     method to be used.
    ///     Because some compression modes represent a set of compression
    ///     methods, this value may be modified to signify the one
    ///     actually choosen.
    /// \param[in] INT_4U Level
    ///     Some compression methods support the idea of levels.
    ///     For these methods, it usually is a trade off between speed
    ///     and size.
    ///     The higher the level, the harder the algorithm tries to
    ///     compress the data.
    /// \param[in] DataType
    ///     The data type of the source.
    /// \param[in] DataTypeMapping
    ///     This is a table mapping the supported data types allowed
    ///     by the user.
    ///     Since these routines are general, they may be able to do
    ///     more than what the caller is requesting.
    ///     This table limits the allowed data types.
    /// \param[in] Source
    ///     This is a buffer of data to be compressed.
    /// \param[in] SourceLength
    ///     The number of data elements in the origional, uncompressed
    ///     data array.
    /// \param[in] SourceByteLength
    ///     The number of bytes used by Source.
    /// \param[out] Dest
    ///     The buffer in compressed format.
    ///     If the compression failed, then this will be a NULL pointer.
    ///     If the compression succeeded, this will be a pointer
    ///     allocated via new[] and is the responsibility of the caller
    ///     to deallocate.
    /// \param[out] DestByteLength
    ///     The number of bytes used by Dest.
    ///     If the compression failed, then this will be zero.
    //-------------------------------------------------------------------
    void
    Compress( INT_4U& CompressionMode, INT_4U Level,
	      const compress_type_mapping& CompressionMapping,
	      INT_4U DataType,
	      const data_type_mapping& DataTypeMapping,
	      const CHAR_U* Source,
	      const INT_8U SourceLength,
	      const INT_8U SourceByteLength,
	      CHAR_U*& Dest,
	      INT_8U& DestByteLength );

    //-------------------------------------------------------------------
    /// \brief uncompress data without altering the source
    ///
    /// \param[in] NatvieOrder
    ///     Boolean describing if the data being uncompressed is of the
    ///     same byte order as the host.
    /// \param[in] CompressionMode
    ///     Version specific compression mode.
    /// \param[in] CompressionMapping
    ///     Map from version specific compression modes to compression
    ///     modes used by the general compression routines.
    /// \param[in] DataType
    ///     Version specific data type.
    /// \param[in] DataTypeMapping
    ///     Map from version specific data type to data types
    ///     used by the general compression routines.
    /// \param[in] Source
    ///     The compressed data stream.
    /// \param[in] SourceLength
    ///     The number of data elements.
    /// \param[in] SourceByteLength
    ///     The number of bytes 
    /// \param[out] Dest
    ///     New buffer containing the decompressed data.
    /// \param[out] DataByteLength
    ///     The number of bytes allocated to Dest.
    //-------------------------------------------------------------------
    void
    Expand( const bool NativeOrder,
	    const INT_4U CompressionMode,
	    const compress_type_mapping& CompressionMapping,
	    const INT_4U DataType,
	    const data_type_mapping& DataTypeMapping,
	    const CHAR_U* Source,
	    const INT_8U SourceLength,
	    const INT_8U SourceByteLength,
	    CHAR_U*& Dest,
	    INT_8U& DataByteLength );

  } // namespace -  Compression
  
} // namespace - FrameCPP

#endif /* FRAME_CPP__Common__Compression_hh */
