#include "framecpp/config.h"

#include <cmath>
#include <sstream>
#include <stdexcept>

#include "general/util.hh"

#include "framecpp/FrameCPP.hh"

#include "framecpp/Common/IOStream.hh"
#include "framecpp/Common/FrameStream.hh"
#include "framecpp/Common/Verify.hh"

#include "framecpp/Version3/FrameSpec.hh"

#include "framecpp/Version3/FrHeader.hh"

using namespace FrameCPP::Version_3;
using FrameCPP::Common::FrameSpec;
using FrameCPP::Common::IStream;
using FrameCPP::Common::OStream;

namespace {
#define	_PI_CONST_ 3.1415926535897932384

  static const INT_2U x1234( 0x1234 );
  static const INT_4U x12345678( 0x12345678 );
  static const INT_8U x123456789abcdef( 0x123456789abcdefLL );

  static const REAL_4 pi_REAL_4( _PI_CONST_ );
  static const REAL_8 pi_REAL_8( _PI_CONST_ );

  template< class T >
  void check_magic_number( const unsigned char* const Source,
			   const T Ref,
			   const std::string& Bytes )
  {
    T buf;

    //-------------------------------------------------------------------
    // Load the buffer
    //-------------------------------------------------------------------
    memcpy(&buf, Source, sizeof( buf ) );
    //-------------------------------------------------------------------
    // Check the bytes
    //-------------------------------------------------------------------
    if ( Ref != buf )
    {
      //-----------------------------------------------------------------
      // Reverse the bytes and check again
      //-----------------------------------------------------------------
      reverse< sizeof( buf ) >( &buf, 1 );
      if ( Ref != buf )
      {
	std::ostringstream	err;

	err << "Data corruption detected in FrHeader at byte(s): "
	    << Bytes;
	//---------------------------------------------------------------
	// Pattern does not match either BIGENDIAN or LITTLEENDIAN so
	//   it must not be a frame file.
	//---------------------------------------------------------------
	throw std::runtime_error( err.str( ) );
      }
    }
  }

  template< class T >
  inline void
  check_pi( const T Source,
	    const T Precision,
	    const std::string& Bytes )
  {
    const T pi( _PI_CONST_ );

    //-------------------------------------------------------------------
    // Check the bytes
    //-------------------------------------------------------------------
    if ( std::abs( Source - pi ) > Precision )
    {
      std::ostringstream	err;

      err << "Value of PI out of range in FrHeader at Bytes: "
	  << Bytes;
      throw std::runtime_error( err.str( ) );
    }
  }
#undef _PI_CONST_
} // namespace - anonymous

namespace FrameCPP
{
  namespace Version_3
  {
    const char* FrHeader::ORIGINATOR_IGWD = "IGWD";

    FrHeader::
    FrHeader( )
      : FrameSpec::Object( s_object_id, (const Common::Description*)NULL )
    {
    }

    FrHeader::
    FrHeader( Common::IStream& Stream )
      : FrameSpec::Object( s_object_id, (const Common::Description*)NULL )
    {
      IStream::char_type* d
	= reinterpret_cast< IStream::char_type* >( &(m_data.raw) );
      Stream.read( d, sizeof( m_data.raw ) );
      //---------------------------------------------------------------------
      // Figure out how to order the bytes
      //---------------------------------------------------------------------
      {
	INT_2U	x;
	memcpy(&x, m_data.raw.x1234, sizeof( x ) );
	m_reorder_bytes = ( x1234 != x );
      }
      if ( m_reorder_bytes )
      {
	if ( FrameCPP::HostByteOrder == FrameCPP::BYTE_ORDER_BIG_ENDIAN )
	{
	  m_byte_order = FrameCPP::BYTE_ORDER_LITTLE_ENDIAN;
	}
	else
	{
	  m_byte_order = FrameCPP::BYTE_ORDER_BIG_ENDIAN;
	}
      }
      else
      {
	m_byte_order = FrameCPP::HostByteOrder;
      }
    }

    FrHeader::
    FrHeader( const FrHeader& Source )
      : FrameSpec::Object( s_object_id, (const Common::Description*)NULL )
    {
      const CHAR_U*	s
	= reinterpret_cast< const CHAR_U* >( &(Source.m_data.raw) );
      CHAR_U*		d
	= reinterpret_cast< CHAR_U* >( &(m_data.raw) );
      std::copy( s, s + sizeof( Source.m_data.raw ), d );
    }

    FrHeader::
    ~FrHeader( )
    {
    }

    FrameSpec::size_type FrHeader::
    Bytes( ) const
    {
      return
	Common::FrHeader::Bytes( )
	+ sizeof( m_data.raw )
	;
    }

    bool FrHeader::
    ByteSwapping( ) const
    {
      return m_reorder_bytes;
    }

    FrameSpec::size_type FrHeader::
    Bytes( const Common::StreamBase& /* Stream */ ) const
    {
      return Bytes( );
    }

    FrHeader* FrHeader::
    Clone( ) const
    {
      return new FrHeader( *this );
    }

    const FrameSpec::Object* FrHeader::
    Description( ) const
    {
      return (const FrameSpec::Object*)NULL;
    }

    FrHeader* FrHeader::
    Create( Common::IStream& Stream ) const
    {
      return new FrHeader( Stream );
    }

#if 0
    void FrHeader::
    Read( Common::IStream& fs )
    {
#if 0
      fs.Read( &m_data, sizeof( m_data ) );

      if ( 0 ) // :TODO: Query the frame stream for level of validation
      {
	//-------------------------------------------------------------------
	// Check the IGWD field
	//-------------------------------------------------------------------
	if ( strcmp( ORIGINATOR_IGWD, m_data.raw.igwd ) != 0 )
	{
	  std::ostringstream	err;
      
	  err << "File originator is not " << ORIGINATOR_IGWD
	      << " (Bytes 0-4 of FrHeader).";
	  std::runtime_error( err.str( ) );
	}
	//-------------------------------------------------------------------
	// Check the ASCII 'A' && 'Z'
	//-------------------------------------------------------------------
	if ( ( m_data.raw.AZ[0] != 'A' ) || ( m_data.raw.AZ[1] != 'Z' ) )
	{
	  std::ostringstream	err;
      
	  err << "File does not support IEEE ASCII hardware standard"
	      << " (Bytes 38-39 of FrHeader).";
	  std::runtime_error( err.str( ) );
	}
      }
      //---------------------------------------------------------------------
      // Check the "magic number". The "magic number" is:
      //   - the three values specifying the byte order
      //---------------------------------------------------------------------

      check_magic_number( m_data.raw.x1234, x1234, "12-13" );
      check_magic_number( m_data.raw.x12345678, x12345678, "14-17" );
      try
      {
	check_magic_number( m_data.raw.x123456789abcdef, x123456789abcdef,
			    "18-25" );
      }
      catch( ... )
      {
	try
	{
	  //-----------------------------------------------------------------
	  // This is a work around for a bug found in some frames where the
	  //   least significant digits were being set to all zeros
	  //-----------------------------------------------------------------
	  static INT_8U	x123456700000000( 0x123456700000000LL );

	  check_magic_number( m_data.raw.x123456789abcdef, x123456700000000,
			      "18-25" );
	}
	catch( ... )
	{
	  //-----------------------------------------------------------------
	  // This is a work around for a bug found in some frames where the
	  //   most significant digits were being set to all zeros
	  //-----------------------------------------------------------------
	  static INT_8U	x0000000001234567( 0x0000000001234567LL );

	  check_magic_number( m_data.raw.x123456789abcdef, x0000000001234567,
			      "18-25" );
	}
      }

      if ( 0 ) // :TODO: Query the frame stream for level of validation
      {
	//-------------------------------------------------------------------
	// Check the values of pi
	//-------------------------------------------------------------------
	REAL_4	pi_4;
	memcpy( &pi_4, m_data.raw.smallPi, sizeof(pi_4) );
	Reorder( pi_4 );
	check_pi( pi_4, REAL_4( 10e-6 ), "26-29" );
	REAL_8	pi_8;
	memcpy( &pi_8, m_data.raw.smallPi, sizeof(pi_8) );
	Reorder( pi_8 );
	check_pi( pi_8, REAL_8( 10e-14 ), "30-37" );
      }
#endif /* 0 */
    }        
#endif /* 0 */

    const char* const FrHeader::
    ObjectStructName( ) const
    {
      static const char* const name = "FrHeader";
      return name;
    }

    void FrHeader::
    VerifyObject( Common::Verify& Verifier,
		  Common::IFrameStream& Stream ) const
    {
      Verifier.ExamineMagicNumber< 2 >( m_data.raw.x1234 );
      Verifier.ExamineMagicNumber< 4 >( m_data.raw.x12345678 );
      Verifier.ExamineMagicNumber< 8 >( m_data.raw.x123456789abcdef );
    }
		  
    void FrHeader::
    Write( Common::OStream& fs ) const
    {
      FrHeader_type	 data;

      Common::FrHeader::Write( fs );
      // Supply the system specific info for the header
      data.raw.sizeInt2 = (unsigned char)sizeof( INT_2U );
      data.raw.sizeInt4 = (unsigned char)sizeof( INT_4U );
      data.raw.sizeInt8 = (unsigned char)sizeof( INT_8U );
      data.raw.sizeReal4 = (unsigned char)sizeof( REAL_4 );
      data.raw.sizeReal8 = (unsigned char)sizeof( REAL_8 );
      memcpy( data.raw.x1234, &x1234, sizeof( x1234 ) );
      memcpy( data.raw.x12345678, &x12345678, sizeof( x12345678 ) );
      memcpy( data.raw.x123456789abcdef, &x123456789abcdef,
	      sizeof( x123456789abcdef ) );
      memcpy( data.raw.smallPi, &pi_REAL_4, sizeof( pi_REAL_4 ) );
      memcpy( data.raw.bigPi, &pi_REAL_8, sizeof( pi_REAL_8 ) );
      data.raw.AZ[ 0 ] = 'A';
      data.raw.AZ[ 1 ] = 'Z';

      // Write the header to the stream
      fs.write( reinterpret_cast< Common::OStream::char_type* >( &data.raw ),
		std::streamsize( sizeof( data.raw ) ) );
    }

    bool FrHeader::
    operator==( const Common::FrameSpec::Object& Obj ) const
    {
      //:TODO: actually implement comparision operator
      throw std::runtime_error( "Comparison operator not supported for FrHeader" );
    }

    std::auto_ptr< FrameSpec::Object > FrHeader::
    demote( INT_2U Target, IStream* Stream ) const
    {
      throw
	Unimplemented( "Object* FrHeader::demote( ) const",
		       DATA_FORMAT_VERSION, __FILE__, __LINE__ );
    }

    std::auto_ptr< FrameSpec::Object > FrHeader::
    promote( INT_2U Target, Object* Obj, IStream* Stream ) const
    {
      throw
	Unimplemented( "Object* FrHeader::promote( ) const",
		       DATA_FORMAT_VERSION, __FILE__, __LINE__ );
    }

  } // namespace - Version_3
} // namespace - FrameCPP
